#include "include.h"

void Engine::init(void *param1, void *param2)
{
#ifdef WIN32
	hwnd = *((HWND *)param1);
	hdc = *((HDC *)param2);
	gfx.init(param1, param2);

#endif

#ifdef __linux__
	display = (Display *)param1;
	window = *((Window *)param2);
#endif

#ifdef __MACH__
	window = *((WindowPtr *)param1);
	agl_context = *((AGLContext *)param2);
#endif

	zNear = 1;
	zFar  = 5001;
	glClearColor(0, 0, 0, 1.0f);
	glClearStencil(0);
	time = 0;

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	glFrontFace(GL_CCW);
	glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_STENCIL_TEST);

	camera_frame.reset();
	camera_frame.pos = vec3(20.0f, 8.0f, 80.0f);
	projection.perspective(45.0, (float)width / height, zNear, zFar, false);


	num_particle = 3000;
	sph.init(num_particle);
}

void Engine::render()
{
	const float identity[16] = {
		1.0f, 0.0f, 0.0f, 0.0f,
		0.0f, 1.0f, 0.0f, 0.0f,
		0.0f, 0.0f, 1.0f, 0.0f,
		0.0f, 0.0f, 0.0f, 1.0f };


	gfx.clear();
	camera_frame.set(transformation);
	matrix4 mvp = transformation * projection;
	glLoadMatrixf(mvp.m);
	sph.render();
	gfx.swap();
	
}

void Engine::dynamics()
{
	cfg_t	config;
	for(unsigned int i = 0; i < entity_list.size(); i++)
	{
		if (entity_list[i]->sleep == true )
			continue;

		RigidBody *body = entity_list[i];
		float delta_time = 0.016f;
		float target_time = delta_time;
		float current_time = 0.0f;
		int divisions = 0;

		if (body->gravity)
			body->net_force = vec3(0.0f, -9.8f * body->mass, 0.0f);

		while (current_time < delta_time)
		{
			body->save_config(config);
			body->integrate(target_time - current_time);
			if ( collision_detect(*body) )
			{
				body->load_config(config);
				target_time = (current_time + target_time) / 2.0f;
				divisions++;

				if (divisions > 10)
				{
//					printf("Entity %d is sleeping\n", i);
					body->sleep = true;
					break;
				}
				continue;
			}
			current_time = target_time;
			target_time = delta_time;
		}
		body->net_force = vec3(0.0f, 0.0f, 0.0f);
	}
}


/*
	Handles all collision detection
	return true if simulated too far.
	return false if collision handled.
*/
bool Engine::collision_detect(RigidBody &body)
{
	Plane floor(vec3(0.0f, 1.0f, 0.0f), 0.0f);
	Plane roof(vec3(0.0f, -1.0f, 0.0f), 10.0f);
	Plane left(vec3(1.0f, 0.0f, 0.0f), 10.0f);
	Plane right(vec3(-1.0f, 0.0f, 0.0f), 10.0f);
	Plane front(vec3(0.0f, 0.0f, 1.0f), 25.0f);
	Plane back(vec3(0.0f, 0.0f, -1.0f), 0.0f);

	if (body_collision(body))
		return true;
	else if (body.collision_detect(floor))
		return true;
	else if (body.collision_detect(roof))
		return true;
	else if (body.collision_detect(left))
		return true;
	else if (body.collision_detect(right))
		return true;
	else if (body.collision_detect(front))
		return true;
	else if (body.collision_detect(back))
		return true;
	else
		return false;
}

//O(N^2)
bool Engine::body_collision(RigidBody &body)
{
	for(unsigned int i = 0; i < entity_list.size(); i++)
	{
		if (entity_list[i] == &body)
			continue;

		if ( body.collision_detect(*entity_list[i]) )
			return true;
	}
	return false;
}

bool Engine::mousepos(int x, int y, int deltax, int deltay)
{
	if (keyboard.leftbutton)
	{
		camera_frame.update(vec2((float)deltax, (float)deltay));
		return true;
	}
	return false;
}

void Engine::keypress(char *key, bool pressed)
{
	char k = 0;

	if (strcmp("enter", key) == 0)
	{
		keyboard.enter = pressed;
	}
	else if (strcmp("leftbutton", key) == 0)
	{
		keyboard.leftbutton = pressed;
		k = 14;
	}
	else if (strcmp("middlebutton", key) == 0)
	{
		keyboard.middlebutton = pressed;
		if (pressed)
		{
			num_particle -= 100;
			if (num_particle < 0)
				num_particle = 0;
			sph.init(num_particle);
		}
		k = 15;
	}
	else if (strcmp("rightbutton", key) == 0)
	{
		keyboard.rightbutton = pressed;
		if (pressed)
		{
			num_particle += 100;
			sph.init(num_particle);
		}

		k = 16;
	}
	else if (strcmp("shift", key) == 0)
	{
		keyboard.shift = pressed;
	}
	else if (strcmp("control", key) == 0)
	{
		keyboard.control = pressed;
	}
	else if (strcmp("escape", key) == 0)
	{
		keyboard.escape = pressed;
	}
	else if (strcmp("up", key) == 0)
	{
		keyboard.up = pressed;
		k = 3;
	}
	else if (strcmp("left", key) == 0)
	{
		keyboard.left = pressed;
		k = 4;
	}
	else if (strcmp("down", key) == 0)
	{
		keyboard.down = pressed;
		k = 5;
	}
	else if (strcmp("right", key) == 0)
	{
		keyboard.right = pressed;
		k = 6;
	}
}

void Engine::step()
{
	camera_frame.update(keyboard);
	sph.step(time);
}


void Engine::resize(int width, int height)
{
	Engine::width = width;
	Engine::height = height;

	//	gfx.resize(width, height);
	glViewport(0, 0, width, height);
	projection.perspective( 45.0, (float) width / height, zNear, zFar, false );
}

void Engine::destroy()
{
	printf("Shutting down.\n");
	gfx.destroy();
//	audio.destroy();
}

void Engine::quit()
{
#ifdef _WINDOWS_
	HWND hwnd = *((HWND *)param1);
	PostMessage(hwnd, WM_CLOSE, 0, 0);
#endif
}

